
!### macro's #####################################################

#define IF_ERROR_RETURN if (status/=0) then; write (gol,'("in ",a)') rname; call goErr; status=1; return; end if

!#################################################################
  
  
module convection

#include "advect.h"

  implicit none

  ! --- interface

  private

  public  ::  Convection_Init, Convection_Done
  public  ::  Convec


  ! --- const --------------------------------------

  character(len=*), parameter  ::  mname = 'convection'
  
  
  ! --- var ---------------------------------------
  
  logical    ::  wet_removal_apply=.true.
  logical    ::  diffusion_apply=.true.


contains



  ! ================================================================
  
  
  subroutine Convection_Init( wet_removal, diffusion, status )
  
    use global_data, only : use_convec
    use Meteo      , only : Set
    use Meteo      , only : entu_dat, entd_dat, detu_dat, detd_dat
    use Meteo      , only : mfw_dat, omega_dat
    use Meteo      , only : cp_sfc
    use Meteo      , only : pu_dat, pv_dat
    use Meteo      , only : temper_dat, humid_dat, gph_dat
    use Meteo      , only : lsmask_sfc
    use Meteo      , only : sr_ecm_sfc, sr_ols_sfc
    use Meteo      , only : u10m_sfc, v10m_sfc
    use Meteo      , only : slhf_sfc, sshf_sfc, ss_sfc


    ! --- in/out --------------------------------
    
    logical, intent(in)            ::  wet_removal
    logical, intent(in)            ::  diffusion
    integer, intent(out)           ::  status
    
    ! --- const ------------------------------

    character(len=*), parameter ::  rname = mname//'/Convection_Init'
    
    ! --- begin --------------------------------
    
    ! convective clouds and vert. diffusion arrays:
    use_convec = .true.
    
    ! entrements/detrements
    call Set( entu_dat, status, used=.true. )
    call Set( entd_dat, status, used=.true. )
    call Set( detu_dat, status, used=.true. )
    call Set( detd_dat, status, used=.true. )

    ! apply wet removal ?
    wet_removal_apply = wet_removal
    ! select meteo if necessary:
    if ( wet_removal_apply ) then
      ! convective precip:
      call Set(   cp_sfc, status, used=.true. )
    end if
    
    ! apply diffusion ?
    diffusion_apply = diffusion

    ! omega is used when creating updr/downdr from raw data;
    ! computed from mfw :
    call Set(   mfw_dat, status, used=.true. )
    call Set( omega_dat, status, used=.true. )
    
    ! ok
    status = 0
    
  end subroutine Convection_Init
  
  
  ! ***
  
  
  subroutine Convection_Done( status )
  
    ! --- in/out --------------------------------
    
    integer, intent(out)           ::  status
    
    ! --- const ------------------------------

    character(len=*), parameter ::  rname = mname//'/Convection_Done'

    ! --- begin --------------------------------
    
    ! nothing to be done

    ! ok
    status = 0
    
  end subroutine Convection_Done
  
  
  
  ! ================================================================
  

  !-----------------------------------------------------------------------
  !
  !****   convec          - mix tracers by convection     v 8.5
  !
  !       programmed by           mh      mpi HH          1-oct-1991
  !       modified by             mh      mpi HH         11-jun-1994
  !
  !       purpose
  !       -------
  !       mix tracers vertically by convection
  !
  !       interface
  !       ---------
  !       call convec
  !
  !       method
  !       ------
  !       the matrix conv is applied simultaneously to rm, rxm and rym. 
  !       The diagonal elements of conv are applied to rzm.
  ! 
  !       externals
  !       ---------
  !       subroutines:    tstamp
  !
  !       reference
  !       ---------       
  !       see manual
  !-----------------------------------------------------------------------
  !       Following the method by Walter Guelle (1997) convective removal of 
  !        tracers is added with some little changes:        
  !       a vector cvsfac is added containing scavenging efficiencies (0-1)
  !       per tracer.
  !                                   aj, knmi may 1998
  !      Maarten Krol, july 2000, implemented zoom version
  !      lmax_conv is now set as the maximum layer to which convection 
  !      is taken into account.
  !
  !       Second moments have been added: Bram Bregman, August 2004
  !--------------------------------------------------------------

  subroutine convec(region)

    use dims,          only : dx, dy, xref, yref, tref
    use dims,          only : isr, ier, jsr, jer
    use dims,          only : im, jm, lm
    use dims,          only : itau, okdebug, kdebug, nconv, lmax_conv
    use dims,          only : adv_scheme, nregions
    use dims,          only : revert
    use global_data,   only : mass_dat,region_dat
    use global_data,   only : emis_data
    use global_data,   only : conv_dat
    use meteo         , only : entu_dat, entd_dat, detu_dat, detd_dat
    use meteo         , only : cp
    use chem_param,    only : ra, ntracet
#ifdef with_budgets    
    use budget_global, only : nzon_vg, buddep_dat, sum_wet, apply_budget_global
    use budget_fine  , only : nzon_v, depwet_cp 
#endif
    use toolbox,       only : lvlpress
    use wet_deposition,only : cvsfac
    use datetime,      only : tstamp
#ifdef MPI
    use mpi_const, only: my_real, &
         mpi_min,mpi_max,com_trac,ierr, ntracet_ar
    use mpi_const, only: pe_first_tracer
#endif
    use ParTools, only : myid, root_t
    use ParTools, only : ntracetloc

    implicit none

    ! input/output
    integer,intent(in) :: region

    ! local
    real,dimension(:,:,:,:), pointer  :: rm,rxm,rym,rzm
#ifdef secmom
    real,dimension(:,:,:,:), pointer  :: rxxm,rxym,rxzm,ryym,ryzm,rzzm
#endif
    real,dimension(:,:,:)  , pointer  :: m
    real,dimension(:,:,:)  , pointer  :: entu, entd, detu, detd, dkg
    integer,dimension(:,:) , pointer  :: cloud_base, cloud_top, cloud_lfs, zoomed
    integer,dimension(3)              :: mnloc
    integer,dimension(3)              :: mxloc
    ! arrays for convection matrix
    ! conv1: convection matrix, lbdcv: wet removal matrix
    real,dimension(lmax_conv,lmax_conv)     :: conv1
    real,dimension(lmax_conv,lmax_conv)     :: lbdcv
    real,dimension(0:lmax_conv,lmax_conv)   :: f
    real,dimension(0:lmax_conv,lmax_conv)   :: fu
    real,dimension(0:lmax_conv)             :: amu,amd
    real,dimension(lmax_conv)               :: new_mass,work
    real,dimension(lmax_conv)               :: zrm,zrxm,zrym,zwetrm
    real,dimension(lmax_conv)               :: srm,srxm,srym,srzm
#ifdef secmom
    real,dimension(lmax_conv)               :: zrxxm,zrxym,zrxzm,zryym,zryzm
    real,dimension(lmax_conv)               :: srxxm,srxym,srxzm,sryym,sryzm,srzzm
#endif
    integer,dimension(lmax_conv)            :: ipvt
    ! thres: threshold in m/s for convective precipitation
    real, parameter                         :: thres = 0.0005/(1e3*3600.*24.)
    real, parameter                         :: area2525 = 2.5*2.5
    real, parameter                         :: area1075 = 10.*7.5
    integer :: n,l,k,j,i
    integer :: imr, jmr, lmr
    integer :: info,ld,li,kk, lshallowtop, nzv
    real    :: scavf,dt,zxi,wetpct
    real    :: minvalue, mxval,mnval,mxvald
    real    :: minvalue_all, mxval_all,mnval_all,mxvald_all
    real    :: scveff, sceffdeep, rtl,area,area1
    integer :: nglob, offsetn

    ! start
    imr=im(region) ; jmr=jm(region) ; lmr=lm(region)

    !WP! only PE's with nonzero ntracet
    if ( ntracetloc == 0 ) return  

#ifdef MPI
    offsetn=sum(ntracet_ar(0:myid-1))
#else
    offsetn=0
#endif

    lshallowtop=lvlpress(region,70000., 98400.)
    ! area in degrees^2, limited in interval...
    area = min(max(dx/xref(region)*dy/yref(region),area2525),area1075)
    sceffdeep = 0.7 + 0.3*(area1075-area)/(area1075-area2525)
    if ( okdebug ) print *, 'convec: Scavenging eff. deep conv. (region=',&
         region,')', sceffdeep

    m => mass_dat(region)%m_t
    rm => mass_dat(region)%rm_t
    rxm => mass_dat(region)%rxm_t
    rym => mass_dat(region)%rym_t
    rzm => mass_dat(region)%rzm_t
#ifdef secmom
    rxxm => mass_dat(region)%rxxm_t
    rxym => mass_dat(region)%rxym_t
    rxzm => mass_dat(region)%rxzm_t
    ryym => mass_dat(region)%ryym_t
    ryzm => mass_dat(region)%ryzm_t
    rzzm => mass_dat(region)%rzzm_t
#endif
    zoomed => region_dat(region)%zoomed
    entu => entu_dat(region)%data
    detu => detu_dat(region)%data
    entd => entd_dat(region)%data
    detd => detd_dat(region)%data
    cloud_top => conv_dat(region)%cloud_top
    cloud_base => conv_dat(region)%cloud_base
    cloud_lfs => conv_dat(region)%cloud_lfs
    dkg => conv_dat(region)%dkg
    
    ! no difusion ? then set diffusion coeff to zero:
    if ( .not. diffusion_apply ) dkg = 0.0

    if ( okdebug ) call tstamp(kdebug,itau,'convec ')
    if ( okdebug ) print *,'convec: Convection called (region)', region

    !timestep and position in packed arrays
    dt = nconv/(2.*tref(region))     

    if ( okdebug ) then
       minvalue = 2.0
       mxvald = -1.0
       mxval = 0.0
       mnval = 2.0
    end if

    ! start main loop over surface cells
    do  j=jsr(region),jer(region)
       do  i=isr(region),ier(region)
          if (zoomed(i,j) /= region) cycle

          ! calculate convection matrix and directly apply to rm(i,j,*,1:ntracet)
          ! Set all amu's, amd's entu' etc to 0 for gridcells without clouds
          !
          ! updraft   - amu is positive upward
          !   
          f(:,:)=0.
          fu(:,:)=0.
          amu(:)=0.
          li = cloud_top(i,j)

          do k=1,li
             amu(k)=amu(k-1)+entu(i,j,k)-detu(i,j,k)
             if ( amu(k) > 0. ) then
                !
                ! we limit zxi from below at 0. in case 
                ! there are inconsistent en/detrainment
                ! rates. The case zxi>1 should not occur
                ! since eu and du are >=0.
                !
                zxi=max(0.,1.-detu(i,j,k)/(amu(k-1)+entu(i,j,k)))

             else
                ! set massflux at upper boundary to strictly 0.

                amu(k)=0.
                zxi=0.
             end if
             do kk=1,k-1
                fu(k,kk)=fu(k-1,kk)*zxi        !all previous levels
             end do
             !cmk: note: the fu(k-1,k) values are zero (see manual Heimann)
             fu(k,k)=entu(i,j,k)/m(i,j,k)*zxi
          end do !k
          !
          ! downdraft  -  amd is negative downward
          !
          amd(:)=0.
          ld = cloud_lfs(i,j)
          do  k=ld,2,-1   
             amd(k-1)=amd(k)-entd(i,j,k)+detd(i,j,k)
             if ( amd(k-1) < 0. ) then
                zxi=max(0.,1.+detd(i,j,k)/(amd(k)-entd(i,j,k)))
             else
                amd(k-1)=0.
                zxi=0.
             end if
             do  kk=k+1,ld   !
                f(k-1,kk)=f(k,kk)*zxi !f is here only fd downdraftmatrix
             end do !kk
             f(k-1,k)=-entd(i,j,k)/m(i,j,k)*zxi    !note the negatives for j  
          end do !k  

          !
          ! add coefficients from up and downdraft together
          !
          do k=1,lmax_conv-1
             do kk=1,lmax_conv
                f(k,kk)=fu(k,kk)+f(k,kk)
             end do !kk
             ! add coefficient from subsidence
             !CMK SH       f(k,k+1)=f(k,k+1)-(amu(k)+amd(k))/m(i,j,k+1)
             ! reported by Sander Houweling.........
             f(k,k+1)=f(k,k+1)-amu(k)/m(i,j,k+1)
             f(k,k  )=f(k,k  )-amd(k)/m(i,j,k  )
             !
             ! diffusion terms
             ! no diffusion in the stratosphere  k>lmax_conv?
             f(k,k)=f(k,k)+dkg(i,j,k)/m(i,j,k)
             f(k,k+1)=f(k,k+1)-dkg(i,j,k)/m(i,j,k+1)
          end do
          
          ! >>> wet removal >>>

          ! initialize wet-removal
          lbdcv(:,:) = 0.0
          
          ! fill scaveging stuff if necessary
          if ( wet_removal_apply ) then

            if ( cp(region)%surf(i,j) > thres ) then
               ! distinguish between deep and shallow convection
               do l=2,lmax_conv
                  if ( l <= lshallowtop ) then
                     scveff=sceffdeep*0.5
                  else
                     scveff=sceffdeep
                  end if
                  lbdcv(l,1:l-1)=scveff
               end do
            end if !cvp(i,j)>thres
            !
            ! WARNING: For matrices f(k,kk), fu(k,kk), and fd(k,kk), index k 
            !          corresponds to the upper boundary of level k 
            !          and index j to level j.
            !          For matrices g(i,j,k,kk) and lbdcv(i,j,k,kk), 
            !          both indices k and j
            !          correspond to respective levels and not to boundaries.
            !
            do k=1,lmax_conv-1
               if ( (f(k,k)-f(k+1,k)) /= 0. ) then
                  wetpct=(fu(k,k)-fu(k+1,k))/(f(k,k)-f(k+1,k))
               else
                  wetpct = 0.0
               end if
               lbdcv(k+1,k)=lbdcv(k+1,k)*wetpct
            end do
            
          end if   ! wet removal ?
          
          ! <<<

          !---- WG end 21/04/98 -----

          ! generate forward matrix
          do k=1,lmax_conv
             do kk=1,lmax_conv
                conv1(k,kk)=-dt*(f(k-1,kk)-f(k,kk))
             end do !kk
             conv1(k,k) = conv1(k,k) + 1.0
          end do !k
          call fastminv(conv1,lmax_conv)

          ! some checks...can be removed later!

          !if(okdebug) then
          !    minvalue = min(minvalue,minval(conv1))
          !    do l=1,lmax_conv
          !        mxval = max(mxval,sum(conv1(:,l)))
          !        mnval = min(mnval,sum(conv1(:,l)))
          !    end do
          !
          !    ! test deviations for a uniform mixing ratio of 1.....ppb
          !
          !    do l=1,lmax_conv
          !       new_mass(l) = dot_product(conv1(l,:),m(i,j,1:lmax_conv)*1e-9)
          !    end do
          !    mxvald = max(mxvald, maxval(abs((m(i,j,1:lmax_conv)*1e-9-new_mass)
          !                 /(m(i,j,1:lmax_conv)*1e-9))))
          !end if   !okdebug
          !
          ! directly apply the matrix:
          ! copy i,j

          if(revert==-1) conv1 = transpose(conv1)

#ifdef MPI   
          do n=1,ntracetloc  !WP! changed to local scope
#else
             do n=1,ntracet
#endif
                nglob=n+offsetn  
                do l=1,lmax_conv
                   srm(l) = rm(i,j,l,n)
                   srxm(l) = rxm(i,j,l,n)
                   srym(l) = rym(i,j,l,n)
                   srzm(l) = rzm(i,j,l,n)
#ifdef secmom
                   srxxm(l) = rxxm(i,j,l,n)
                   srxym(l) = rxym(i,j,l,n)
                   srxzm(l) = rxzm(i,j,l,n)
                   sryym(l) = ryym(i,j,l,n)
                   sryzm(l) = ryzm(i,j,l,n)
                   srzzm(l) = rzzm(i,j,l,n)
#endif
                   zwetrm(l) = 0.0 
                   zrm(l)=0.  
                   zrxm(l) = 0.0
                   zrym(l) = 0.0
#ifdef secmom
                   zrxxm(l) = 0.0
                   zrxym(l) = 0.0
                   zrxzm(l) = 0.0
                   zryym(l) = 0.0
                   zryzm(l) = 0.0
#endif
                end do
                do l=1,lmax_conv
                   srzm(l)=conv1(l,l)*srzm(l)   !rzm is directly calculated
#ifdef secmom
                   srzzm(l)=conv1(l,l)*srzzm(l)   !rzzm is directly calculated
#endif
                   do k=1,lmax_conv
                      ! species dependent removal....
                      scavf=lbdcv(l,k)*cvsfac(nglob)
                      zrm(l)=zrm(l)+(1.0-scavf)*conv1(l,k)*srm(k)
                      zwetrm(l) = zwetrm(l)+scavf*conv1(l,k)*srm(k)
                      zrxm(l)=zrxm(l)+conv1(l,k)*srxm(k)
                      zrym(l)=zrym(l)+conv1(l,k)*srym(k)
#ifdef secmom
                      zrxxm(l)=zrxxm(l)+conv1(l,k)*srxxm(k)
                      zrxym(l)=zrxym(l)+conv1(l,k)*srxym(k)
                      zrxzm(l)=zrxzm(l)+conv1(l,k)*srxzm(k)
                      zryym(l)=zryym(l)+conv1(l,k)*sryym(k)
                      zryzm(l)=zryzm(l)+conv1(l,k)*sryzm(k)
#endif
                   end do    !k
                end do      !l
                do l=1,lmax_conv
                   rm(i,j,l,n)=zrm(l)
                   rxm(i,j,l,n)=zrxm(l)
                   rym(i,j,l,n)=zrym(l)
                   rzm(i,j,l,n)=srzm(l)
#ifdef secmom
                    rxxm(i,j,l,n)=zrxxm(l)
                    rxym(i,j,l,n)=zrxym(l)
                    rxzm(i,j,l,n)=srxzm(l)
                    ryym(i,j,l,n)=zryym(l)
                    ryzm(i,j,l,n)=zryzm(l)
                    rzzm(i,j,l,n)=srzzm(l)
#endif
                end do
#ifdef with_budgets    
                if ( apply_budget_global ) then
                  if (cvsfac(nglob).gt.0.) then
                     do l=1,lmax_conv
                        nzv=nzon_vg(l)
                        buddep_dat(region)%cp(i,j,nzv,nglob) = &
                             buddep_dat(region)%cp(i,j,nzv,nglob) + &
                             zwetrm(l)/ra(nglob)*1.e3       !kg=> moles
                        if ( region == nregions ) then
                           nzv=nzon_v(l)
                           depwet_cp(i,j,nzv,nglob) = depwet_cp(i,j,nzv,nglob) + &
                                zwetrm(l)/ra(nglob)*1.e3  
                        end if
                        if ( nglob == 1 ) sum_wet(region) = sum_wet(region) + &
                             zwetrm(l)  !in kg!
                     end do !l
                  end if
                end if
#endif

             end do          !ntracet
          end do   !i
       end do     !j

       !if (okdebug)  then
       !
       !   call mpi_allreduce(minvalue,minvalue_all,1,my_real, &
       !        mpi_min,com_trac,ierr)
       !   call mpi_allreduce(mxval,mxval_all,1,my_real,mpi_max,com_trac,ierr)
       !   call mpi_allreduce(mnval,mnval_all,1,my_real,mpi_min,com_trac,ierr)
       !   call mpi_allreduce(mxvald,mxvald_all,1,my_real,mpi_max,com_trac,ierr)
       !   call barrier_t
       !   if (myid==root_t) then
       !    print* ,'-----convection matrix information----------------------'
       !    print* ,'Minimum value of convection matrix :',minvalue_all
       !    print *,'Max/Min along the columns  (?/=1)  :',mxval_all,mnval_all  
       !     !these should be one/zero to ensure mass-conservation...
       !    print *,'Maximum fractional deviation from uniform mixing ', &
       !       'ratio column:',mxvald_all  
       !    print* ,'------convection matrix information---------------------'
       !   end if
       !   call barrier_t
       !   if (mxvald_all > 0.1) &
       !      call exitus(1, ' Convection matrix appears wrong') 
       !end if

       nullify(m)
       nullify(rm)
       nullify(rxm)
       nullify(rym)
       nullify(rzm)
#ifdef secmom
       nullify(rxxm)
       nullify(rxym)
       nullify(rxzm)
       nullify(ryym)
       nullify(ryzm)
       nullify(rzzm)
#endif
       nullify(zoomed)
       nullify(entu)
       nullify(detu)
       nullify(entd)
       nullify(detd)
       nullify(cloud_top)
       nullify(cloud_base)
       nullify(cloud_lfs)
       nullify(dkg)

     end subroutine convec


#ifdef with_lapack

     subroutine fastminv(a,lm)
       !-----------------------------------------------------------------------
       ! compute inverse matrix using LU decompositon and forward-backward
       ! substitution without pivoting (for easy vectorization)
       !
       ! Source: EMR and NR, sec. edition (p. 35ff).
       !
       !                                        mh, {Sat, Jun 4, 1994} 16:27:15 
       !                                        LAPACK version, WP, Jan-2006
       !-----------------------------------------------------------------------

       use toolbox,       only: escape_tm
       implicit none

       ! input/output
       integer,intent(in)                  :: lm
       real,dimension(lm,lm),intent(inout) :: a

       ! local
       integer :: istat
       real,dimension(lm) :: work
       real,dimension(lm) :: ipiv

       call DGETRF( lm, lm, a, lm, ipiv, istat ) 
       if(istat.ne.0) call escape_tm('fastminv: Failed to make LU decomposition ')
       call DGETRI( lm, a, lm, ipiv, work, lm, istat )
       if(istat.ne.0) call escape_tm('fastminv: Failed to make inverse matrix')

     end subroutine fastminv

#else

     subroutine fastminv(a,lm)
       !-----------------------------------------------------------------------
       ! compute inverse matrix using LU decompositon and forward-backward
       ! substitution without pivoting (for easy vectorization)
       !
       ! Source: EMR and NR, sec. edition (p. 35ff).
       !
       !                                        mh, {Sat, Jun 4, 1994} 16:27:15 
       !-----------------------------------------------------------------------

       implicit none

       ! input/output
       integer,intent(in)                  :: lm
       real,dimension(lm,lm),intent(inout) :: a

       ! local
       real,dimension(lm)    :: y
       real,dimension(lm)    :: b
       real,dimension(lm,lm) :: bb
       integer :: i,j,k

       do j=1,lm

          do i=1,j
             bb(i,j)=a(i,j)
             do k=1,i-1
                bb(i,j)=bb(i,j)-bb(i,k)*bb(k,j)
             end do
          end do

          do i=j+1,lm
             bb(i,j)=a(i,j)
             do k=1,j-1
                bb(i,j)=bb(i,j)-bb(i,k)*bb(k,j)
             end do
             bb(i,j)=bb(i,j)/bb(j,j)
          end do
       end do

       do k=1,lm

          do i=1,lm
             b(i)=0.
          end do
          b(k)=1.
          do i=1,lm
             y(i)=b(i)
             do j=1,i-1
                y(i)=y(i)-bb(i,j)*y(j)
             end do
          end do

          do i=lm,1,-1
             a(i,k)=y(i)
             do j=i+1,lm
                a(i,k)=a(i,k)-bb(i,j)*a(j,k)
             end do
             a(i,k)=a(i,k)/bb(i,i)
          end do

       end do


     end subroutine fastminv

#endif


   end module convection
